#include "setplay_message_builder.h"

#include <sstream>
#include <cstdlib>
#include <typeinfo>

#include "setplay_message_parser.h"

#include <rcsc/common/server_param.h>
#include <rcsc/common/audio_codec.h>
#include <rcsc/common/logger.h>

#include <setplay/communication/messages.h>
#include <setplay/generators/sexpr/setplay_generator_sexpr.h>

using namespace std;
using namespace rcsc;


/*
 * SaySetplayMessage
 */

std::ostream& SaySetplayMessage::printBase( const setplay::IMessage& msg,
                                            std::ostream & os ) const
{
    os << "SetplayId=" << msg.setplayId()
       << " Invertible=" << (msg.inverted() ? "TRUE" : "FALSE");

    return os;
}

std::ostream& SaySetplayMessage::printArguments( const setplay::IMessage& msg,
                                                 std::ostream & os ) const
{
    os << "Arguments: ";
    if ( !msg.arguments().empty() )
    {
        setplay::SetplayGeneratorSexpr sos( os );
        std::map<std::string, setplay::Object*>::const_iterator it = msg.arguments().begin();
        while ( it != msg.arguments().end() )
        {
            os << " ";
            it->second->visit( sos );
            it++;
        }
        os << " )]";
    }
    else
    {
        os << " none";
    }
    return os;
}

void SaySetplayMessage::encodeSetplayMessageBase( const setplay::IMessage& msg, std::string& to ) const
{
    static unsigned int max_setplays = 32; // 5 bits for set-plays

    /// MessageBitSchema: SetplayId Inverted Arguments
    /// FirstByte: 00SSSSSI
    /// Remaining bytes: tuple of punum
    /// Septlay
    /// Inverted (1bits): SSSS
    /// PlayerUnum (4bits): PPPP
    assert ( msg.setplayId() < max_setplays );

    // 1) Encode set-play id => This should be further abstracted
    // Setplay-Id => Must use one byte or agree on the number of used set-plays => Must do renunmbering (no way to infer which plan id is used if mixed with other info)
    unsigned int res = 0;

    res = msg.setplayId();
    res <<= 1;

    // 2) Encode set-play invertible flag
    res |= msg.inverted() ? 1 : 0;

    to += AudioCodec::i().intToCharMap().at(res);
}

void SaySetplayMessage::encodeSetplayMessageArguments( const setplay::IMessage& msg,
                                                       std::string& to ) const
{
    if ( msg.arguments().empty() )
    {
        return;
    }

    /// Max numbers that can be encoded in a single byte (order matters for set-play instantiation, cannot change without losing info)

    // Unum => Multiples of 11
    setplay::IMessage::Arguments::const_iterator it_end = msg.arguments().end();
    setplay::IMessage::Arguments::const_iterator it = msg.arguments().begin();

    while( it != it_end )
    {
        setplay::Object* object = it->second;

        switch( object ->type() )
        {
            case setplay::TT_PLAYER_AFFILIATION:
            {
                // TODO Assuming only our players
                const setplay::PlayerAffiliation* player = static_cast<const setplay::PlayerAffiliation*>( object  );
                to += AudioCodec::i().unum2hex( player->number() );
            }
            break;
            default:
                // TODO Handle only player affiliations for now
                assert(false);
            break;
        }

        it++;
    }
}

unsigned int SaySetplayMessage::lengthSetplayMessageBase(const setplay::IMessage&  ) const
{
    return 1; // 1 byte for set-play id and inverted state
}

unsigned int SaySetplayMessage::lengthSetplayMessageArguments(const setplay::IMessage& msg) const
{
    if ( msg.arguments().empty() )
    {
        return 0;
    }

    // HACK Assuming that arguments are only unums (1..16) each occupying 4 bits (half a byte)
    // return ceil(msg.arguments().size() / 2);
    return msg.arguments().size();
}


 /*
  * SaySetplayStartMessage
  */

SaySetplayStartMessage::SaySetplayStartMessage( const setplay::SetplayStartMessage *msg )
    : M_msg( msg )
{

}

SaySetplayStartMessage::~SaySetplayStartMessage()
{
    delete M_msg;
}

char SaySetplayStartMessage::header() const
{
    return SetplayStartMessageParser::sheader();
}

int SaySetplayStartMessage::length() const
{
    unsigned int msg_size = 0;

    msg_size += 1; // The char for the header
    msg_size += lengthSetplayMessageBase( *M_msg );
    msg_size += lengthSetplayMessageArguments( *M_msg );

    return msg_size;
}

bool SaySetplayStartMessage::toStr (std::string &to) const
{
    if ( (int)to.length() + length() > ServerParam::i().playerSayMsgSize() )
    {
        dlog.addText( Logger::SENSOR,
                      "InitMessage. over the message size : buf = %d, this = %d",
                      to.length(), length() );
        return false;
    }

    dlog.addText(Logger::SENSOR, "Trying to create a SetplayStartMessage string");

    std::string msg_str;

    // 1) Encode set-play id and invertibility
    encodeSetplayMessageBase( *M_msg, msg_str );

    // 2) Encode all arguments ( if any )
    encodeSetplayMessageArguments( *M_msg, msg_str );

    to += header();
    to += msg_str;

    dlog.addText( Logger::SENSOR,
                  __FILE__ " SetplayStartMessage. success! %s",
                  to.c_str() );
    return true;
}

std::ostream & SaySetplayStartMessage::printDebug( std::ostream & os ) const
{
    os << "[SaySetplayStartMessage ";

    printBase( *M_msg, os );
    printArguments( *M_msg, os );

    os << " ]";

    return os;
}

/*
 * StepTransitionMessage
 */
SayStepTransitionMessage::SayStepTransitionMessage( const setplay::StepTransitionMessage *msg )
    : M_msg(msg)
{

}

SayStepTransitionMessage::~SayStepTransitionMessage()
{
    delete M_msg;
}

char SayStepTransitionMessage::header() const
{
    return StepTransitionMessageParser::sheader();
}

int SayStepTransitionMessage::length() const {
    unsigned int msg_size = 0;

    msg_size += 1; // The char for the header
    msg_size += lengthSetplayMessageBase( *M_msg );
    msg_size += lengthSetplayMessageArguments( *M_msg );

    /// TODO Encode step ids (4 bits for each is more than enough) in a single char
    // msg_size += 1; // Include current step id and next step id as 1 hex char
    msg_size += 2;

    return msg_size;
}

bool SayStepTransitionMessage::toStr (std::string &to) const
{
    if ( (int)to.length() + length() > ServerParam::i().playerSayMsgSize() )
    {
        dlog.addText( Logger::SENSOR,
                      "StepTransitionMessage. over the message size : buf = %d, this = %d",
                      to.length(), length() );
        return false;
    }

    std::string msg = "";

    dlog.addText(Logger::SENSOR, "Started the creation of a StepTransitionMessage string");

    // 1) Encode set-play id and invertible flag
    encodeSetplayMessageBase( *M_msg, msg );

    // 2) Encode all arguments ( if any )
    encodeSetplayMessageArguments( *M_msg, msg );

    // 3) Encode current step id and next step id
    encodeSteps( msg );

    to += header();
    to += msg;

    dlog.addText( Logger::SENSOR,
                  "StepTransitionMessage. success!"
                 " setplayId=%d inverted=%d from=%d to=%d -> [%s]",
                  M_msg->setplayId(),
                  M_msg->inverted(),
                  M_msg->currentStepId(),
                  M_msg->nextStepId(),
                  to.c_str());
    return true;
}

void SayStepTransitionMessage::encodeSteps( std::string& to ) const
{
    /*unsigned int res = 0;
    res |= M_msg->currentStepId();
    res <<=4; // Use for bits for each step id
    res |= M_msg->nextStepId();

    assert( res < (unsigned int) AudioCodec::i().CHAR_SIZE );

    to += AudioCodec::i().intToCharMap().at( res );*/
    to += AudioCodec::i().intToCharMap().at( M_msg->currentStepId() );
    to += AudioCodec::i().intToCharMap().at( M_msg->nextStepId() );
}

std::ostream & SayStepTransitionMessage::printDebug( std::ostream & os ) const
{
    os << "[SayStepTransitionMessage: ";

    printBase( *M_msg, os );
    printArguments( *M_msg, os );

    os << " fromStep: " << M_msg->currentStepId()
       << " toStep: " << M_msg->nextStepId();

    os << " ]";
    return os;
}

/*
 * StepEnterMessage
 */
SayStepEnterMessage::SayStepEnterMessage( const setplay::StepEnterMessage *msg )
    : M_msg(msg)
{

}

SayStepEnterMessage::~SayStepEnterMessage()
{
    delete M_msg;
}

char SayStepEnterMessage::header() const
{
    return StepEnterMessageParser::sheader();
}

int SayStepEnterMessage::length() const
{
    unsigned int msg_size = 0;
    msg_size += 1; // The char for the header
    msg_size += lengthSetplayMessageBase( *M_msg );
    msg_size += lengthSetplayMessageArguments( *M_msg );
    msg_size += 1; // Include current step id as 1 hex char (Assume to be sufficient)

    return msg_size;
}

bool SayStepEnterMessage::toStr (std::string &to) const
{
    if ( (int)to.length() + length() > ServerParam::i().playerSayMsgSize() )
    {
        dlog.addText( Logger::SENSOR,
                      "StepEnterMessage. over the message size : buf = %d, this = %d",
                      to.length(), length() );
        return false;
    }

    dlog.addText(Logger::SENSOR, "Started the creation of a StepEnterMessage string");

    std::string msg = "";

    // 1) Encode set-play id and invertible
    encodeSetplayMessageBase( *M_msg, msg );

    // 2) Encode set-play arguments
    encodeSetplayMessageArguments( *M_msg, msg );

    // 3) Encode current step id
    msg += AudioCodec::i().intToCharMap().at( M_msg->stepId() );

    to += header();
    to += msg;

    dlog.addText( Logger::SENSOR,
                  "SayStepEnterMessage. success!"
                 " setplayId=%d inverted=%d step=%d -> [%s]",
                  M_msg->setplayId(),
                  M_msg->inverted(),
                  M_msg->stepId(),
                  to.c_str());

    return true;
}

std::ostream & SayStepEnterMessage::printDebug( std::ostream & os ) const
{
    os << "[SayStepEnterMessage ";

    printBase( *M_msg, os );
    printArguments( *M_msg, os );

    os << " step: " << M_msg->stepId();

    os << " ]";
    return os;
}

/*
 * SetplayAbortMessage
 */

SaySetplayAbortMessage::SaySetplayAbortMessage( const setplay::SetplayAbortMessage* msg )
    : M_msg( msg )
{

}

char SaySetplayAbortMessage::header() const
{
   return SetplayAbortMessageParser::sheader();
}

int SaySetplayAbortMessage::length() const
{
    unsigned int msg_size = 0;

    msg_size += 1; // Header char
    msg_size += lengthSetplayMessageBase( *M_msg ); // Setplay id

    return msg_size;
}

bool SaySetplayAbortMessage::toStr (std::string &to) const
{
    if ( (int)to.length() + length() > ServerParam::i().playerSayMsgSize() )
    {
        dlog.addText( Logger::SENSOR,
                      "AbortMessage. over the message size : buf = %d, this = %d",
                      to.length(), length() );
        return false;
    }

    dlog.addText(Logger::SENSOR, "Trying to create a SetplayAbortMessage string");

    std::string msg;

    // 1) Encode set-play id and invertible
    encodeSetplayMessageBase( *M_msg, msg );

    to += header();
    to += msg;

    dlog.addText( Logger::SENSOR,
                  "SaySetplayAbortMessage. success!"
                 " setplayId=%d inverted=%d -> [%s]",
                  M_msg->setplayId(),
                  M_msg->inverted(),
                  to.c_str());


    return true;
}

std::ostream & SaySetplayAbortMessage::printDebug( std::ostream & os ) const
{
    os << "[SaySetplayAbortMessage ";

    printBase( *M_msg, os);

    os << ']';
    return os;
}


/*
 * SetplayFinishMessage
 */

SaySetplayFinishMessage::SaySetplayFinishMessage( const setplay::SetplayFinishMessage* msg )
    : M_msg( msg )
{

}

char SaySetplayFinishMessage::header() const
{
   return SetplayFinishMessageParser::sheader();
}

int SaySetplayFinishMessage::length() const
{

    unsigned int msg_size = 0;

    msg_size += 1; // Header char
    msg_size += lengthSetplayMessageBase( *M_msg );

    return msg_size;
}


bool SaySetplayFinishMessage::toStr (std::string &to) const
{
    if ( (int)to.length() + length() > ServerParam::i().playerSayMsgSize() )
    {
        dlog.addText( Logger::SENSOR,
                      "FinishMessage. over the message size : buf = %d, this = %d",
                      to.length(), length() );
        return false;
    }

    dlog.addText(Logger::SENSOR, "Trying to create a SetplayFinishMessage string");

    std::string msg = "";

    // 1) Encode set-play id => This should be further abstracted
    encodeSetplayMessageBase( *M_msg, msg );

    to += header();
    to += msg;

    dlog.addText( Logger::SENSOR,
                  "SaySetplayFinishMessage. success!"
                 " setplayId=%d inverted=%d -> [%s]",
                  M_msg->setplayId(),
                  M_msg->inverted(),
                  to.c_str());

    return true;
}

std::ostream & SaySetplayFinishMessage::printDebug( std::ostream & os ) const
{
    os << "[SaySetplayFinishMessage ";

    printBase( *M_msg, os );

    os << ']';

    return os;
}

/*
 * ActionDoneMessage
 */

/* bool ActionDoneMessage::toStr (std::string &to) const
{

}

std::ostream & ActionDoneMessage::printDebug( std::ostream & os ) const
{
    os << "[SetplayActionDone: at " << emission_time() << ']';
} */

