#include "setplay.h"

#include <setplay/model/transitions/next_step.h>
#include <setplay/model/conditions/icondition.h>
#include <setplay/model/geom/iregion.h>

#include <cassert>
#include <algorithm>

using namespace std;

namespace setplay {

    Setplay::Setplay()
        : M_id(0)
        , M_name("undefined")
        , M_invertible(false)
        , M_abort_condition(NULL)
        , M_wait_condition(NULL)
        , M_start_condition(NULL)
    {
        M_steps = new Steps();
        M_parameters = new Parameters();
        M_players =new Participants();
        M_assignments = new Assignments();
    }

    // TODO Copy constructor (Unfinished => Don't use this)
    Setplay::Setplay(const Setplay& other)
        : M_id( other.M_id )
        , M_name( other.M_name )
        , M_invertible(other.M_invertible)
    {
        M_start_condition = other.M_start_condition != NULL
                                ? other.M_start_condition->clone()
                                : NULL;

        M_abort_condition = other.M_abort_condition != NULL
                                ? other.M_abort_condition->clone()
                                : NULL;

        M_wait_condition = other.M_wait_condition != NULL
                             ? other.M_wait_condition->clone()
                             : NULL;

        M_steps = new Steps();
        if ( other.M_steps != NULL )
        {
            for ( Steps::iterator its = other.M_steps->begin();
                  its != other.M_steps->end();
                  its++)
            {
                addStep( its->second->clone() );
            }
        }

        M_players = new Participants();
        if ( other.M_players != NULL  )
        {
            for ( Participants::iterator itp = other.M_players->begin();
                  itp != other.M_players->end();
                  itp++)
            {
                addParticipant( (*itp)->clone() );
            }
        }

        M_parameters = new Parameters();
        if ( other.M_parameters != NULL )
        {
            for ( Parameters::iterator itp = other.M_parameters->begin();
                  itp != other.M_parameters->end();
                  itp++ )
            {
                addParameter( itp->second->clone() );
            }
        }

        M_assignments = new Assignments();
        if ( other.M_assignments != NULL )
        {
            for ( Assignments::iterator ita = other.M_assignments->begin();
                  ita != other.M_assignments->end();
                  ita++)
            {
                addAssignment( (*ita)->clone() );
            }
        }

        /// Calculate execution paths
    }

    Setplay::~Setplay()
    {
        // Reclaim space
        if ( M_start_condition != NULL )
        {
            delete M_start_condition;
        }

        if ( M_wait_condition != NULL )
        {
            delete M_wait_condition;
        }

        if ( M_abort_condition != NULL )
        {
            delete M_abort_condition;
        }

        if ( M_steps != NULL )
        {
            for ( Steps::iterator its = M_steps->begin();
                  its != M_steps->end();
                  its++)
            {
                delete its->second;
            }
            delete M_steps;
        }

        if ( M_players != NULL  )
        {
            for ( Participants::iterator itp = M_players->begin();
                  itp != M_players->end();
                  itp++)
            {
                delete *itp;
            }
            delete M_players;
        }

        if ( M_parameters != NULL )
        {
            for ( Parameters::iterator itp = M_parameters->begin();
                  itp != M_parameters->end();
                  itp++)
            {
                delete itp->second;
            }
            delete M_parameters;
        }

        if ( M_assignments != NULL )
        {
            for ( Assignments::iterator ita = M_assignments->begin();
                  ita != M_assignments->end();
                  ita++)
            {
                delete *ita;
            }
            delete M_assignments;
        }

    }

    Step* Setplay::getStep(unsigned int stepID) const
    {
        Steps::iterator its = M_steps->find( stepID );
        assert ( its != M_steps->end() );

        return its->second;
    }

    const vector<PlayerAffiliation*>& Setplay::playerIds() const
    {
        return M_player_ids;
    }

    const vector<VarPlayer*>& Setplay::playerVars() const
    {
        return M_player_vars;
    }

    void Setplay::splitParticipants()
    {
        if ( firstStep()->numParticipations() == 0 )
        {
            return;
        }

        const Participations* first_step_participations = firstStep()->participations();

        Participations::const_iterator it = first_step_participations->begin();
        while ( it++ != first_step_participations->end() )
        {
            IPlayer* player= (*it)->player();
            switch ( player->type() )
            {
                case TT_PLAYER_AFFILIATION:
                    M_player_ids.push_back( static_cast<PlayerAffiliation*>(player) );
                    break;
                case TT_VAR_PLAYER:
                    M_player_vars.push_back( static_cast<VarPlayer*>(player) );
                    break;
            default:
                cerr << "Wrong type of data " << player->type() << endl;
                assert(false);
            }
        }
    }

    unsigned int Setplay::id() const
    {
        return M_id;
    }

    string Setplay::name() const
    {
        return M_name;
    }

    bool Setplay::invertible() const
    {
        return M_invertible;
    }

    Steps* Setplay::steps() const
    {
        return M_steps;
    }

    Parameters* Setplay::parameters() const
    {
        return M_parameters;
    }

    Assignments* Setplay::assignments() const
    {
        return M_assignments;
    }

    Participants Setplay::initialPlayers() const
    {
        return firstStep()->participants();
    }

    // SETPLAY EXECUTION PATHS
    // TODO This is not plan-implementation specific
    Step* Setplay::firstStep() const
    {
        return getStep(Step::FIRST_STEP_ID);
    }

    // TODO This is not soccer specific !! Should not be implemented here
    vector<Step *> Setplay::getTerminationSteps()
    {
        vector<Step*> end_steps;
        for (Steps::const_iterator it_step = M_steps->begin();
             it_step != M_steps->end();
             it_step++)
        {
            Step* step = it_step->second;
            if ( step->transitions()->empty() )
            { // SoccerStep has no transitions
                end_steps.push_back( step );
            }
            else
            {
                for(Transitions::const_iterator it_transition = step->transitions()->begin();
                    it_transition != step->transitions()->end();
                    it_transition++)
                {
                    if (!(*it_transition)->type() == TT_TRANSITION_NEXT_STEP)
                    { // SoccerStep as at least 1 finish/abort transition
                        end_steps.push_back( step );
                        break;
                    }
                }
            }
        }
        return end_steps;
    }

    vector<ExecutionPath*>* Setplay::executionPaths() const
    {
        return executionPathsFrom( firstStep() );
    }

    vector<ExecutionPath*>* Setplay::executionPathsFrom(Step* src_step) const
    {
        vector<ExecutionPath*>* paths = new vector<ExecutionPath*>();
        ExecutionPath* path = new ExecutionPath();

        executionPaths(paths, path, src_step);

        return paths;
    }

    vector<ExecutionPath*>* Setplay::executionPathsTo(Step* dst_step) const
    {
        return executionPaths(firstStep(), dst_step);
    }

    vector<ExecutionPath*>* Setplay::executionPaths( Step *src_step,
                                                     Step *dst_step) const
    {
        vector<ExecutionPath*>* paths = new vector<ExecutionPath*>();
        ExecutionPath* path = new ExecutionPath();

        executionPaths(paths, path, src_step, dst_step);

        return paths;
    }

    ICondition* Setplay::startCondition() const
    {
        return M_start_condition;
    }

    ICondition* Setplay::abortCondition() const
    {
        return M_abort_condition;
    }

    ICondition* Setplay::waitCondition() const
    {
        return M_wait_condition;
    }

    bool Setplay::hasSteps( ) const
    {
        return M_steps != NULL && !M_steps->empty();
    }

    bool Setplay::hasParameters( ) const
    {
        return M_parameters != NULL && !M_parameters->empty();
    }

    bool Setplay::hasAssignments( ) const
    {
        return M_assignments != NULL && !M_assignments->empty();
    }


    bool Setplay::hasStartCondition() const
    {
        return M_start_condition != NULL;
    }

    bool Setplay::hasAbortCondition() const
    {
        return M_abort_condition != NULL;
    }

    bool Setplay::hasWaitCondition() const
    {
        return M_wait_condition != NULL;
    }

    Participants* Setplay::players() const
    {
        return M_players;
    }

    /// Modifiers with fluent interfaces
    Setplay& Setplay::setId(unsigned int id)
    {
        M_id = id;
        return *this;
    }

    Setplay& Setplay::setName(std::string name)
    {
        M_name = name;
        return *this;
    }

    Setplay& Setplay::setInvertible(bool invertable)
    {
        M_invertible = invertable;
        return *this;
    }

    Setplay& Setplay::setStartCondition(ICondition* condition)
    {
        if ( M_start_condition != NULL ) { delete M_start_condition; }
        M_start_condition = condition;
        return *this;
    }

    Setplay& Setplay::setWaitCondition(ICondition* condition)
    {
        if ( M_wait_condition != NULL ) { delete M_wait_condition; }
        M_wait_condition = condition;
        return *this;
    }

    Setplay& Setplay::setAbortCondition(ICondition* condition)
    {
        if ( M_abort_condition != NULL ) { delete M_abort_condition; }
        M_abort_condition = condition;
        return *this;
    }

    Setplay& Setplay::setPlayers(Participants* players)
    {
        if ( M_players != NULL )
        {
            for ( Participants::iterator it = M_players->begin();
                  it != M_players->end(); it++ )
            {
                delete *it;
            }
            delete M_players;
            M_player_vars.clear();
            M_player_ids.clear();
        }
        M_players = players;
        return *this;
    }

    Setplay& Setplay::setParameters(Parameters* parameters)
    {
        if ( M_parameters != NULL )
        {
            for ( Parameters::iterator it = M_parameters->begin();
                  it != M_parameters->end(); it++ )
            {
                delete it->second;
            }
            delete M_parameters;
        }
        M_parameters = parameters;
        // TODO Must reset playerIds and playerVarss
        return *this;
    }

    Setplay& Setplay::setSteps(Steps* steps)
    {
        if ( M_steps != NULL )
        {
            for ( Steps::iterator it = M_steps->begin();
                  it != M_steps->end(); it++ )
            {
                delete it->second;
            }
            delete M_steps;
        }
        M_steps = steps;
        return *this;
    }

    Setplay& Setplay::setAssignments(Assignments* assignments)
    {
        if ( M_assignments != NULL )
        {
            for ( Assignments::iterator it = M_assignments->begin();
                  it != M_assignments->end(); it++)
            {
                delete *it;
            }
            delete M_assignments;
        }
        M_assignments = assignments;
        return *this;
    }

    Setplay& Setplay::addStep(Step* step)
    {
        M_steps->insert(pair<unsigned int, Step*>( step->id(), step));
        return *this;
    }

    Setplay& Setplay::addParticipant(IPlayer* participant)
    {
        M_players->push_back(participant);
        return *this;
    }

    Setplay& Setplay::addParameter(Variable* parameter)
    {
        M_parameters->insert(pair<string, Variable*>(parameter->name(), parameter));
        return *this;
    }

    Setplay& Setplay::addAssignment(Assignment* assignment)
    {
        M_assignments->push_back(assignment);
        return *this;
    }


    void Setplay::executionPaths(vector<ExecutionPath*>* paths,
                                    ExecutionPath* path,
                                    Step* src_step) const
    {
        path->addStep( src_step);

        // Have we found a step without transitions
        if ( src_step->numTransitions() == 0)
        {
            paths->push_back(path); // We are at the end
            return;
        }

        bool finish_transition_found = false; // No need to add the path twice
        // Continue the path search in the transitions to the next steps
        Transitions::const_iterator it_transition = src_step->transitions()->begin();
        if ( (*it_transition)->type() == TT_TRANSITION_NEXT_STEP )
        {
            NextStep* next_step_transition = static_cast<NextStep*>( *it_transition );
            Step* next_step = getStep( next_step_transition->nextStepId() );
            const vector<Step*>& steps = path->steps();
            if ( find(steps.begin(), steps.end(), next_step) == steps.end() )
            { // Next step isn't already in the list
                executionPaths(paths, path, next_step);
            }
            else
            {
                paths->push_back( path );
            }
        }
        else
        { // The cur step has Finish or Abort Transitions => If it has several, the path should only be added once
            finish_transition_found = true;
            paths->push_back(path);
        }


        while ( ++it_transition != src_step->transitions()->end())
        {
            if ((*it_transition)->type() == TT_TRANSITION_NEXT_STEP)
            {
                NextStep* next_step_transition = static_cast<NextStep*>( *it_transition );
                Step* next_step = getStep( next_step_transition->nextStepId() );
                const vector<Step*>& steps = path->steps();
                if (find(steps.begin(), steps.end(), next_step) == steps.end())
                { // Next step isn't already in the list (TODO This verification isn't ok)
                    ExecutionPath* new_path = new ExecutionPath(*path);
                    executionPaths(paths, new_path, next_step);
                }
            }
            else
            { // The current step also has Finish or Abort Transitions => If it has several, the path should only be added once
                if (!finish_transition_found)
                {
                    ExecutionPath* new_path = new ExecutionPath(*path);
                    paths->push_back(new_path);
                }
                finish_transition_found = true;
            }
        }
    }

    void Setplay::executionPaths(vector<ExecutionPath*>* paths_list,
                                    ExecutionPath* path,
                                    Step *src_step,
                                    Step *dst_step) const
    {
        path->addStep( src_step );

        // Have we reached the end
        if (src_step->id() == dst_step->id())
        {
            paths_list->push_back(path);
            return;
        }

        bool path_found = false;
        // Continue the path search in the transitions to the next steps
        for (Transitions::const_iterator it_transition = src_step->transitions()->begin();
            it_transition != src_step->transitions()->end();
            it_transition++)
        {
            if ((*it_transition)->type() == TT_TRANSITION_NEXT_STEP)
            {
                NextStep* next_step_transition = static_cast<NextStep*>( *it_transition );
                Step* next_step = getStep( next_step_transition->nextStepId() );
                const vector<Step*>& steps = path->steps();
                if (find(steps.begin(), steps.end(), next_step) == steps.end())
                { // Next step isn't already in the list
                    ExecutionPath* new_path = path;
                    executionPaths(paths_list, new_path, next_step, dst_step);
                }
                else if (!path_found)
                { // The current step also has Finish or Abort Transitions => If it has several, the path should only be added once
                    path_found = true;
                    paths_list->push_back(path);
                }
            }
        }
        if (!path_found)
        {
            delete path; // Clean allocated memory
        }
    }

    Setplay* Setplay::clone() const
    {
        return new Setplay(*this);
    }

    void Setplay::visit(IVisitor& visitor)
    {
        visitor.visit( *this );
    }

}

