// -*-c++-*-
#ifndef SP_SETPLAY_H
#define SP_SETPLAY_H

#include <setplay/model/object.h>

#include <map>
#include <string>
#include <iostream>

#include <setplay/model/step.h>
#include <setplay/model/conditions/icondition.h>
#include <setplay/model/geom.h>
#include <setplay/model/participations/participation_situated.h>

#include <setplay/model/scope.h> // NOTE Scope is builtin in parameters !!!!
#include <setplay/model/execution_path.h>
#include <setplay/model/assign.h>

using namespace std;

namespace setplay {

    typedef std::map<unsigned int, Step*> Steps;
    typedef std::map<std::string, Variable*> Parameters;

    /// TODO
    /// Add invertibility check condition => Can an estimator be used????
    /// Add set-play success general condition
    /// Add step-entry actions??

    class Setplay
            : public Object {
    public:

        Setplay();

        // Copy constructor
        Setplay(const Setplay& other);

        ~Setplay();

        inline TokenType type() const
        {
          static TokenType type = TT_SETPLAY;
          return type;
        }

        // Inspectors
        unsigned int id() const;
        string name() const;
        Participants *players() const;
        Steps* steps() const;
        Parameters* parameters() const;
        Assignments* assignments() const;
        bool invertible() const;

        const vector<PlayerAffiliation*>& playerIds() const;
        const vector<VarPlayer*>& playerVars() const;

        ICondition* startCondition() const;
        ICondition* abortCondition() const;
        ICondition* waitCondition() const;

        bool hasSteps( ) const;
        bool hasParameters( ) const;
        bool hasAssignments( ) const;
        bool hasStartCondition( ) const;
        bool hasAbortCondition( ) const;
        bool hasWaitCondition( ) const;

        /// Modifiers with fluent interfaces
        Setplay& setId(unsigned int id);
        Setplay& setName(std::string name);
        Setplay& setInvertible(bool invertible);
        Setplay& setStartCondition(ICondition* condition);
        Setplay& setWaitCondition(ICondition* condition);
        Setplay& setAbortCondition(ICondition* condition);
        Setplay& setPlayers(Participants* players);
        Setplay& setParameters(Parameters* parameters);
        Setplay& setSteps(Steps* steps);
        Setplay& setAssignments(Assignments* assignments);
        Setplay& addStep(Step* step);
        Setplay& addParticipant(IPlayer* participant);
        Setplay& addParameter(Variable* parameter);
        Setplay& addAssignment(Assignment* assignment);

        /// Overload unary '-' operator
        /// Setplay* operator-() const;

        Step *firstStep() const;

        vector<Step*> getTerminationSteps();

        /**
        * \brief Get all possible execution paths from the start step to all existing end steps
        *
        * @return Set of all possible execution paths
        */
        vector<ExecutionPath* >* executionPaths() const;

        /**
        * \brief Get all execution paths start from the first step and end in any destination step
        *       * @param srcStep The step at which the execution step shall end
        * @return Set of all possible execution paths that start from the source step
        */
        vector<ExecutionPath* >* executionPathsFrom( Step* src_step ) const;

        /**
        * \brief Get all execution paths that end in dstStep and start from the first step
        *
        * @param dstStep The step at which the execution step shall end
        * @return Set of all possible execution paths from the first step to the specific destination step
        */
        vector<ExecutionPath* >* executionPathsTo( Step *dst_step ) const;

        /**
        * \brief Get all execution paths starting from srcStep and ending in dstStep
        *
        * @param srcStep The step from which the execution path will start
        * @param dstStep The step at which the execution step shall end
        * @return Set of all possible execution paths between the source and destination step
        */
        vector<ExecutionPath* >* executionPaths( Step* src_step,
                                                 Step* dst_step) const;

        Step* getStep(unsigned int step_id) const;

        Setplay* clone() const;
        void visit(IVisitor& visitor);

        protected:
            Participants initialPlayers() const;

            // 14/06/11 FA - Can the current player update the status of the setplay
            bool canUpdateCurrentStepStatus(const IContextProvider& observer);

            // FA 10/06 Get relevant players for the player (omit this logic from the agent)
            vector<const PlayerAffiliation*>* getRelevantPlayers();

            // FA 23/07/11 - Setplay execution paths
            void executionPaths( vector<ExecutionPath* >* paths,
                                 ExecutionPath* path,
                                 Step* src_step,
                                 Step* dst_step) const;
            void executionPaths(vector<ExecutionPath* >* paths,
                                 ExecutionPath* path,
                                 Step *src_step) const;

            //Attributes

        private:
            void splitParticipants();

            unsigned int M_id; // May be negative, for inverted setplays
            string M_name;
            bool M_invertible;
            Steps* M_steps;
            Participants* M_players;
            ICondition* M_abort_condition;
            ICondition* M_wait_condition;
            ICondition* M_start_condition;
            string M_comment;
            string M_version;

            vector<VarPlayer*> M_player_vars;
            vector<PlayerAffiliation*> M_player_ids;

            Parameters* M_parameters;
            Assignments* M_assignments;

            /// Steps calculus
            std::vector<Step*> M_end_steps;
            std::vector< ExecutionPath > M_execution_paths;
    };

    /*friend ostream &operator<<(ostream &out, Setplay& setplay);
    friend ostream &operator<<(ostream &out, Step& setplay);
    friend ostream &operator<<(ostream &out, Transition& setplay);
    friend ostream &operator<<(ostream &out, Directive& setplay);
    friend ostream &operator<<(ostream &out, IRegion& setplay);
    friend ostream &operator<<(ostream &out, IAction& setplay);
    friend ostream &operator<<(ostream &out, Cond& setplay);*/

}
#endif //SETPLAY_H
