/*
 * File:   icontext_provider.h
 * Author: falmeida
 *
 * Created on May 31, 2011, 6:35 PM
 */

#ifndef SP_WORLD_MODEL_H
#define	SP_WORLD_MODEL_H

#include <string>
#include <rcsc/geom.h>

#include <setplay/model/participants/player_affiliation.h>
#include <setplay/model/game_situation.h>

#include <rcsc/types.h>

namespace setplay {

    class IWorldModel
    {
    public:
        virtual ~IWorldModel() { }

        virtual unsigned int currentTime() const = 0;
        virtual unsigned int remainingTimeInPeriod() const = 0;

        virtual unsigned int ballInterceptionTime( const PlayerAffiliation& player) const = 0;
        virtual unsigned int playerInterceptionTime( const PlayerAffiliation& player1,
                                                     const PlayerAffiliation& player2 ) const = 0;

        // Game stopped?
        virtual GameSituation gameSituation() const = 0;

        /*!
          \brief Determine the current time period
          */
        virtual TimePeriod timePeriod() const = 0;

        /// COMMON REGIONS
        virtual rcsc::Rect2D field( ) const = 0;
        virtual double defenseLine(TeamAffiliation team_affiliation) const = 0;
        virtual double offenseLine(TeamAffiliation team_affiliation) const = 0;
        double offsideLine(TeamAffiliation team_affiliation) const
        {
            assert ( team_affiliation != TA_UNKNOWN );
            return defenseLine( team_affiliation == TA_OUR ? TA_OPP : TA_OUR );
        }

        virtual double goalLine(TeamAffiliation team_affiliation) const = 0;
        virtual double endLine(TeamAffiliation team_affiliation) const = 0;

        virtual rcsc::Rect2D goalieArea(TeamAffiliation team_affiliation) const = 0;
        virtual rcsc::Rect2D penaltyArea(TeamAffiliation team_affiliation) const = 0;
        virtual rcsc::Rect2D midfieldArea(TeamAffiliation team_affiliation) const = 0;

        /// BALL INFORMATION
        virtual rcsc::Vector2D ballPosition( ) const = 0;
        virtual double ballSpeed() const = 0;
        virtual rcsc::AngleDeg ballDirection() const = 0;

        /// PLAYER INFORMATION
        virtual unsigned int self() const = 0;
        virtual unsigned int goalie( TeamAffiliation team_affiliation ) const = 0;
        virtual rcsc::Vector2D playerPosition( const PlayerAffiliation& player ) const = 0;
        virtual double playerSpeed( const PlayerAffiliation& player ) const = 0;
        virtual rcsc::AngleDeg playerBodyDirection( const PlayerAffiliation& player ) const = 0;
        virtual int playerTimeToReachPosition( const PlayerAffiliation& affiliation,
                                               const rcsc::Vector2D& position ) const = 0;
        virtual int playerTimeToReachMaxSpeed( const PlayerAffiliation& player ) const = 0;
        virtual double playerMaxSpeed( const PlayerAffiliation& player ) const = 0;


        virtual std::vector<unsigned int> activePlayers( TeamAffiliation team_affiliation ) const = 0;

        virtual PlayerAffiliation playerFastestTo( const rcsc::Vector2D& position ) = 0;
        virtual PlayerAffiliation playerFastestTo( const rcsc::Region2D& region ) = 0;
        virtual PlayerAffiliation playerFastestTo( const PlayerAffiliation& player ) = 0;
        virtual PlayerAffiliation playerFastestToBall( ) = 0;

        virtual unsigned int teammateFastestTo( const rcsc::Vector2D& position ) = 0;
        virtual unsigned int teammateFastestTo( const rcsc::Region2D& region ) = 0;
        virtual unsigned int teammateFastestTo( const PlayerAffiliation& player ) = 0;
        virtual unsigned int teammateFastestToBall( ) = 0;

        virtual unsigned int opponentFastestTo( const rcsc::Vector2D& position ) = 0;
        virtual unsigned int opponentFastestTo( const rcsc::Region2D& region ) = 0;
        virtual unsigned int opponentFastestTo( const PlayerAffiliation& player ) = 0;
        virtual unsigned int opponentFastestToBall( ) = 0;

        virtual std::vector<unsigned int> teammatesFromPlayer( const PlayerAffiliation& player ) const = 0;
        virtual std::vector<unsigned int> opponentsFromPlayer( const PlayerAffiliation& player ) const = 0;

        virtual std::vector<unsigned int> playersFromBall( TeamAffiliation team_affiliation ) const = 0;
        virtual std::vector<unsigned int> playersFromGoal( TeamAffiliation team_affiliation,
                                                           TeamAffiliation goal_affiliation) const = 0;
        virtual std::vector<unsigned int> playersBehindBall( TeamAffiliation team_affiliation) const = 0;
        virtual std::vector<unsigned int> playersInFrontOfBall( TeamAffiliation team_affiliation) const = 0;

        // virtual std::vector<IPlayerId*> teammatesToPlayer( const IPlayerId& player_id ) const = 0;
        // virtual std::vector<IPlayerId*> teammatesToPosition( const rcsc::Vector2D& position ) const = 0;
        virtual PlayerAffiliation lastKicker() const = 0;

        virtual TeamAffiliation teamInPossession() const = 0;


        virtual unsigned int goalsScored(TeamAffiliation affiliation) const = 0;

        int relativeScore() const
        {
            return goalsScored( TA_OUR ) - goalsScored( TA_OPP );
        }

        /*!
          \brief Determine the players belonging to a particular team affiliation that are offside
          \param team_affiliation The team affiliation the players belong to
          \returns Players that are offside
          */
        virtual std::vector<unsigned int> playersOffside(TeamAffiliation team_affiliation) const = 0;
        /*!
          \brief Determine the players belonging to a particular team affiliation that are not offside
          \param team_affiliation The team affiliation the players belong to
          \returns Players that are not offside
          */
        virtual std::vector<unsigned int> playersNotOffside(TeamAffiliation team_affiliation) const = 0;

    };

}


#endif	/* SP_ICONTEXT_PROVIDER_H */
