#include "setplay_augmenter.h"

/// Needed for StepWaitConditionAugmenter
#include <setplay/model/conditions/cond_and.h>
#include <setplay/model/conditions/cond_object_pos.h>

/// Needed for VirtualTransitionAugmenter
#include <setplay/model/actions/act_pos.h>
#include <setplay/model/actions/act_dribble.h>
#include <setplay/model/actions/act_pass.h>
#include <setplay/model/directives/do.h>
#include <setplay/model/transitions/next_step.h>
#include <setplay/model/conditions.h>

namespace setplay {

StepWaitConditionAugmenter::StepWaitConditionAugmenter(bool ignore_steps_with_wait_conditions)
    : M_ignore_steps_with_wait_conditions( ignore_steps_with_wait_conditions )
{

}

bool StepWaitConditionAugmenter::augment( Setplay* setplay )
{
    bool augmented = false;

    for ( Steps::iterator its = setplay->steps()->begin() ;
          its != setplay->steps()->end() ; its++ )
    {
        // Only augment if it has no wait condition
        Step* step = its->second;

        if ( step->hasWaitCondition() && M_ignore_steps_with_wait_conditions )
        {
            // Ignore steps already with wait conditions
            continue;
        }

        /// TODO If M_ignore_steps_with_wait_conditions then we should try not to add duplicate conditions
        ICondition* wait_condition = generateStepWaitConditionFromParticipations( step );
        if ( wait_condition != NULL  )
        {
            step->setWaitCondition( wait_condition );
            augmented = true;
        }
    }

    return augmented;
}

ICondition* StepWaitConditionAugmenter::generateStepWaitConditionFromParticipations( const Step* step )
{
    /// Derive a condition from the step participations
    CondAnd* cond_step_participants_locations = new CondAnd();
    for ( Participations::const_iterator itp = step->participations()->begin();
          itp != step->participations()->end(); itp++ )
    {
        const IParticipation* step_participation = *itp;
        if ( step_participation->hasLocation() )
        {
            CondAtLocation* cond_participant_at_location = new CondAtLocation( step_participation->player()->clone(),
                                                                   step_participation->location()->clone() );
            cond_step_participants_locations->push( cond_participant_at_location );
        }
    }

    if ( cond_step_participants_locations->conditions()->empty() )
    {
        delete cond_step_participants_locations;
        return NULL;
    }
    CondNot* wait_condition = new CondNot( cond_step_participants_locations );
    return wait_condition;
}

VirtualTransitionsAugmenter::VirtualTransitionsAugmenter(bool check_bowner_action )
    : M_check_bowner_action( check_bowner_action  )
{

}

bool VirtualTransitionsAugmenter::augment( Setplay* setplay )
{
    if ( setplay->steps()->size() < 2 )
    {
        return false;
    }

    bool augmented = false;

    Steps::iterator last_step_it = setplay->steps()->end();
    last_step_it--;
    Steps::iterator penultimate_step_it = last_step_it;
    penultimate_step_it--;

    for ( Steps::iterator its1 = setplay->steps()->begin() ; its1 != penultimate_step_it ; its1++ )
    {
        Steps::iterator its2 = its1;
        its2++;
        for ( ; its2 != last_step_it ; its2++ )
        {
            /// If there is already a transition between both steps do not create a new transition
            Step* step1 = its1->second;
            Step* step2 = its2->second;
            if ( !step1->getTransitionToStep( step2->id() ) )
            {
                /// Should a virtual transition be created between both steps?
                augmented = augmented || createTransition( setplay, step1, step2);
            }

            if ( !step2->getTransitionToStep( step1->id() ) )
            {
                /// Should a virtual transition be created between both steps?
                augmented = augmented || createTransition( setplay, step2, step1);
            }

        }
    }

    return augmented;
}

bool VirtualTransitionsAugmenter::createTransition( const Setplay* setplay,
                                                          Step* from,
                                                          const Step *to)
{
    NextStep* next_step = new NextStep();
    next_step->setNextStepId( to->id() );

    const IParticipation* bowner_participation_from_step = stepLeaderParticipation( from );
    const IParticipation* bowner_participation_to_step = stepLeaderParticipation( to );

    for ( Participations::const_iterator itpt = from->participations()->begin();
          itpt != from->participations()->end() ; itpt++ )
    {
        const IParticipation* player_next_step_participation = playerStepParticipation( (*itpt)->player(), to );
        if ( player_next_step_participation != NULL &&
             player_next_step_participation->hasLocation() &&
             !samePlayer( bowner_participation_from_step->player(), (*itpt)->player() ) )
        {
            /// Decide player actions => For now just do a pos to the next step initial location
            ActPos* act_pos = new ActPos( player_next_step_participation->location()->clone() );
            Do* do_participant = new Do();
            do_participant ->addPlayer( (*itpt)->player()->clone() );
            do_participant ->addAction( act_pos );
            next_step->addDirective( do_participant );
        }
    }

    /// TODO Should use a player comparator!!!
    if ( samePlayer( bowner_participation_from_step->player(), bowner_participation_to_step->player() ) )
    {
        if ( bowner_participation_to_step->hasLocation() )
        {
            /// Generate a dribble action
            ActDribble* act_dribble = new ActDribble( bowner_participation_to_step->location()->clone() );
            Do* do_dribble_bowner = new Do();
            do_dribble_bowner ->addPlayer( bowner_participation_from_step->player()->clone() );
            do_dribble_bowner ->addAction( act_dribble );
            next_step->addDirective( do_dribble_bowner );

            if ( M_check_bowner_action )
            {
                // TODO next_step->setStartCondition( new CondCanDribble( act_dribble->location()->clone() ) );
                assert( false );
            }
        }
    }
    else
    {
        /// Generate a pass action from current ball owner to new step ball owner
        ActPass* act_pass = new ActPass( new PlayerPosition( bowner_participation_to_step->player()->clone() ) );
        Do* do_pass_bowner = new Do();
        do_pass_bowner ->addPlayer( bowner_participation_from_step->player()->clone() );
        do_pass_bowner ->addAction( act_pass );
        next_step->addDirective( do_pass_bowner );

        if ( M_check_bowner_action )
        {
            next_step->setStartCondition( new CondCanPass( bowner_participation_from_step->player()->clone(),
                                                           act_pass->location()->clone() ) );
        }
    }

    /// Add the directive to the next step
    from->addTransition( next_step );

    return true;
}

const IParticipation* VirtualTransitionsAugmenter::playerStepParticipation( const IPlayer* player, const Step* step )
{

    for ( Participations::const_iterator itp = step->participations()->begin();
          itp != step->participations()->end() ; itp++ )
    {
        /// Check if player participates in a step
        const IParticipation* step_participation;
        if ( samePlayer( player, step_participation->player() ))
        {
            return step_participation;
        }

    }

    return NULL;
}

bool VirtualTransitionsAugmenter::samePlayer( const IPlayer* p1, const IPlayer* p2)
{
    /// If the types don't match say no
    /// TODO What if the player has been assigned to a variable prior to reaching the step
    switch ( p1->type() )
    {
        case TT_PLAYER_AFFILIATION:
            const PlayerAffiliation* p1_affil = static_cast<const PlayerAffiliation*>( p1 );
            switch ( p2->type() )
            {

            }
        break;
    }
}

const IParticipation *VirtualTransitionsAugmenter::stepLeaderParticipation( const Step* step )
{
    /// If the step has no ball owner, return NULL
    /// HACK For now assume that it is the step leader. In the future look for a ball ownership condition and see who it applies to
    return playerStepParticipation(step->leader(), step);
}

/// SetplayGeneratorAbortConditionAugmenter
OffensiveSetplayAbortConditionAugmenter::OffensiveSetplayAbortConditionAugmenter()
{

}

bool OffensiveSetplayAbortConditionAugmenter::augment( Setplay* setplay )
{
    CondTeamInPossession* opp_team_in_possession = new CondTeamInPossession( TA_OPP );

    // Check if there is already an abort condition in the set-play
    if ( !setplay->hasAbortCondition() )
    {
        setplay->setAbortCondition( opp_team_in_possession );
        return true;
    }

    if ( setplay->abortCondition()->type() == TT_CONDITION_OR )
    {
        CondOr* cond_or = static_cast<CondOr*>( setplay->abortCondition() );
        /// TODO Should check if CondOr does not already evaluate to this condition to prevent cluttering
        cond_or->push( opp_team_in_possession );
        return true;
    }

    /// Add a condOr to accomodate the existing conditions, without loss of generality
    /// TODO Should check if the existing condition does not already checks this to prevent cluttering
    CondOr* cond_or = new CondOr();
    cond_or->push( setplay->abortCondition()->clone() );
    cond_or->push( opp_team_in_possession );
    return true;
}

}
