#include "setplay_selector.h"

#include <cstdlib>

#include <rcsc/common/logger.h>

using namespace std;

namespace setplay {

FirstFeasibleSetplaySelector::FirstFeasibleSetplaySelector()
{

}

bool FirstFeasibleSetplaySelector::selectSetplay( SetplayInterpreter& interpreter )
{
    if ( M_database->count() < 0 )
    {
        std::cerr << __FILE__ << " No setplays to choose from " << std::endl;
    }

    for( Setplays::const_iterator its = M_database->setplays().begin();
         its != M_database->setplays().end();
         its++)
    {
        const Setplay* setplay = its->second;

        interpreter.loadSetplay( setplay );

        if ( setplay->hasParameters() )
        {
            // Instantiate setplay
             if ( !M_player_selector->selectPlayers( interpreter ) )
             {
                 rcsc::dlog.addText( rcsc::Logger::PLAN,
                                     __FILE__ " Could not allocate all necessary players! Reset set-play interpreter");
                 interpreter.reset();
                 continue;
             }
        }

        // TODO Check if it is invertable and test both situations
        // Invertability might affect startCondition evaluation outcome
        if ( setplay->hasStartCondition() &&
             interpreter.interpret( *setplay->startCondition() ) )
        {
            ///  Set-play start condition is TRUE
            rcsc::dlog.addText( rcsc::Logger::PLAN,
                                __FILE__ " Selected setplay %d due to SetplayStartCondition",
                                setplay->id());
            return true;
        }

        const Step* first_step = setplay->firstStep();
        if ( !first_step->hasStartCondition() ||
             ( first_step->hasStartCondition() && interpreter.interpret( *first_step->startCondition() )))
        {
            ///  First step start condition is TRUE
            rcsc::dlog.addText( rcsc::Logger::PLAN,
                                __FILE__ " Selected setplay %d due to FirstStepStartCondition",
                                setplay->id());
            return true;
        }

        interpreter.reset();
    }
    rcsc::dlog.addText( rcsc::Logger::PLAN,
                        __FILE__ " No setplay could be selected from %d",
                        M_database->count());
    return false;
}

void FirstFeasibleSetplaySelector::setPlayerSelector( IPlayerSelector *player_selector )
{
    M_player_selector = player_selector;
}

void FirstFeasibleSetplaySelector::setDatabase( const SetplayDatabase* database )
{
    M_database = database;
}

/// CbrSetplaySelector

/* const float CbrSetplaySelector::S_first_evaluation = 100.0;

CbrSetplaySelector::CbrSetplaySelector()
{

}

Setplay* CbrSetplaySelector::selectSetplay(const SetplayMediator& sm )
{
    std::map<unsigned int, double> setplays_evaluation;

    const SetplayDatabase& sdb = sm.setplayDatabase();
    double sum_evaluations = 0.0;

    // Calculate set-plays evaluation
    for ( Setplays::const_iterator it = sdb.setplays().begin();
          it !=  sdb.setplays().end(); it++)
    {
        double setplay_evaluation = evaluate( *it->second);
        sum_evaluations += setplay_evaluation;
        setplays_evaluation.insert(pair<unsigned int, double>( it->first,
                                                               setplay_evaluation));
    }

    // Choose the setplay by a random number between 0 and sumEvaluations:
    // this way the evaluation scores will be the relative probability of
    // a setplay being chosen
    double randDouble = rand() * sum_evaluations / RAND_MAX;

    //DEBUG
    std::cerr << "SP: CBR choosing SP, finding "
              << randDouble << " in "
              << sum_evaluations
              << " size:" << sdb.setplays().size()
              << endl;

    double accumulated_evaluations = 0;
    for ( Setplays::const_iterator it = sdb.setplays().begin();
          it != sdb.setplays().end(); it++ )
    {
        accumulated_evaluations += setplays_evaluation.at(it->first);
        if ( accumulated_evaluations >= randDouble )
        {
            return it->second;
        }
    }

    return NULL;
}

double CbrSetplaySelector::evaluate(const Setplay& setplay)
{

    CaseBase::Cases* game_cases = M_game_case_base.getSetplayCases( setplay.id() );
    CaseBase::Cases* past_cases = M_historical_case_base.getSetplayCases( setplay.id() );
    if ( game_cases == NULL || past_cases == NULL ) // Award standard value for Setplays with no registered executions
    {
        return S_first_evaluation;
    }

    // Evaluate setplay in the current game
    float game_cases_evaluation = 0.0;

    // Historical evaluation of the set-play
    float past_cases_evaluation = 0.0;

    M_case_evaluator->evaluate( *it );

    double thisGameEvalSum = 0,
           thisGameWeightSum = 0,
           thisTeamEvalSum = 0,
           thisTeamWeightSum = 0,
           otherTeamsEvalSum = 0,
           otherTeamsWeightSum = 0;

    for( CaseBase::Cases::const_iterator itc = game_cases->begin();
         itc != game_cases->end(); itc++ )
    {
          Case::Evaluation thisEval = it->second.at(j)->evaluate( world );

          //DEBUG
          cerr << "SP: CBR case " << possibleSetplays.at(i).name
               << " eval:" << thisEval
               << endl;

          thisGameEvalSum += thisEval.evalThisGame * thisEval.weightThisGame;
          thisGameWeightSum += thisEval.weightThisGame;
          thisTeamEvalSum += thisEval.evalThisTeam * thisEval.weightThisTeam;
          thisTeamWeightSum += thisEval.weightThisTeam;
          otherTeamsEvalSum += thisEval.evalOtherTeams*thisEval.weightOtherTeams;
          otherTeamsWeightSum += thisEval.weightOtherTeams;
    }

    // In this game, the eval can be 0, since the case may refer to tries
    // with other opponents
    totalEval
          = (thisGameWeightSum == 0 ? 0 : thisGameEvalSum / thisGameWeightSum )
          * weightThisGame
          + (thisTeamWeightSum == 0 ? 0 : thisTeamEvalSum / thisTeamWeightSum )
          * weightThisTeam
          + (otherTeamsWeightSum == 0 ? 0 : otherTeamsEvalSum / otherTeamsWeightSum)
          * weightOtherTeams;
    return totalEval;
}

// Lmota
// To make distance increase with separation, index difference must be
// calculated to some power, originally 4 for longitudinal and transversal
// and 2 for radial. I thought this was too high, and changed to 2 and 1.5.
unsigned int
CbrSetplaySelector::longitudinalRegionDistance(const std::string& reg1,
                                               const std::string& reg2)
{
    return regionsDistance( S_longitudinal_regions, reg1, reg2, 2);
}

unsigned int
CbrSetplaySelector::transversalRegionsDistance(const std::string& reg1,
                                             const std::string& reg2)
{
    return regionsDistance( S_transversal_regions, reg1, reg2, 2);
}

unsigned int
CbrSetplaySelector::radialRegionsDistance(const std::string& reg1,
                                         const std::string& reg2)
{
    return regionsDistance( S_radial_regions, reg1, reg2, 1.5);
}

unsigned int
CbrSetplaySelector::regionsDistance( RegionPositions& regions,
                                     const std::string& region1,
                                     const std::string& region2,
                                     double exponent)
{
    return (unsigned int)pow( (double) regions.at(region1) -
                                       regions.at(region2),
                              exponent);
} */

}
