#include "strategy.h"

#include <soccer/iworld_model.h>
#include <soccer/player_id.h>

#include <strategy/tactic/tactic_with_situations.h>

#include <strategy/formation/formation_sbsp.h>
#include <strategy/formation/formation_fcp_dt_2d.h>

#include <strategy/flux/flux_matrix_2d.h>
#include <strategy/flux/flux_dt.h>

#include <strategy/strategy_config.h>
#include <strategy/flux_config.h>
#include <strategy/formation_config.h>
#include <strategy/setplay_config.h>
#include <strategy/league_config.h>
#include <strategy/tactic_config.h>

using namespace soccer;
using namespace soccer::tactic;
using namespace soccer::formation;
using namespace soccer::flux;

using namespace soccer::strategy::util;

namespace soccer {

namespace strategy {

    Strategy::Strategy()
        : M_current_tactic( NULL )
        , M_current_formation( NULL )
        , M_last_formation( NULL )
        , M_current_flux( NULL )
    {

    }


    Strategy::~Strategy()
    {

    }


    bool Strategy::init( const string& strategy_config_file )
    {
        bool success = true;

        // TODO Implement fcportugal Strategy::init
        // HACK
        read( string("strategy.conf.2D") );

        StrategyConfig& SC = StrategyConfig::instance();
        success &= SC.init(strategy_config_file);

        if (!success)
        {
            cerr << __FILE__ << " Error importing strategy configuration file " << endl;
            return false;
        }

        // Setplays config
        if ( !SC.setplaysConfigFile().empty())
        {
            success &= SetplayConfig::instance().init(SC.setplaysConfigFile());
        }

        // Tactics config
        if ( !SC.tacticsConfigFile().empty())
        {
            success &= TacticConfig::instance().init(SC.tacticsConfigFile());
        }

        // Formations config
        if ( !SC.formationsConfigFile().empty())
        {
            success &= FormationConfig::instance().init(SC.formationsConfigFile());
        }

        // Fluxes config
        if ( !SC.fluxesConfigFile().empty())
        {
            success &= FluxConfig::instance().init(SC.fluxesConfigFile());
        }

        // League config
        if ( !SC.leagueConfigFile().empty())
        {
            success &= LeagueConfig::instance().init(SC.leagueConfigFile());
        }

        // Player types
        if ( !SC.playerTypesConfigFile().empty())
        {
            cerr << __FILE__ << " Error: Must implement player types config" << endl;
        }

        return true; // success
    }


    const Vector2D& Strategy::position(const soccer::Unum& number) const
    {
        return M_positions.at(number);
    }

    /*
    bool Strategy::shouldApplyDPRE() const
    {
        // Goalie does not execute DPRE in 2D
        return M_dpre->active() && StrategyConfig::instance().useDPRE();
    }*/


    void Strategy::update(  const IWorldModel *world_model  )
    {
        updateTactic( world_model );

        updateFormation ( world_model );

        updatePositions( world_model );
    }

    /*************************************/
    /*** Strategy Configuration Import ***/
    /*************************************/


    bool Strategy::readTimeTactics(istream& is,
                                   unsigned int num_time_tactics,
                                   unsigned int num_opponent_tactics)
    {
         TimeTacticSelector* time_tactic_selector = new TimeTacticSelector( );

        cout << "Reading " << num_time_tactics << " TimeTactics" << endl;

        for(unsigned int tt=0; tt < num_time_tactics; tt++)
        {
            time_tactic_selector->read( is, num_opponent_tactics );
        }

        M_tactic_selector = time_tactic_selector;
        return true;
    }


    bool Strategy::readTactics(istream& is, unsigned int num_tactics)
    {
        cout << "Reading " << num_tactics << " Tactics" << endl;

        for(unsigned int k=0; k < num_tactics; k++)
        {
            Tactic* tactic = new TacticWithSituations();
            if ( !tactic->read(is) )
            {
                delete tactic;
                return false;
            }
            M_tactics.insert(Tactics::value_type(tactic->id(), tactic ));
        }
        return true;
    }


    bool Strategy::readPlayerTypesDefinitions(istream& is,
                                              unsigned int num_player_types)
    {
        cout << "Reading " << num_player_types << " Player Types" << endl;

        istringstream istr(getNextLine(is));

        unsigned int player_roles[num_player_types];
        double attraction_x[num_player_types];
        double attraction_y[num_player_types];
        double area_att_att[num_player_types];
        double area_att_def[num_player_types];
        int behind_ball[num_player_types];
        double pos_min_x[num_player_types];
        double pos_max_x[num_player_types];
        double pos_min_y[num_player_types];
        double pos_max_y[num_player_types];
        int decision_algorithm[num_player_types+1];

        for(unsigned int k=1; k <= num_player_types; k++)
        {
            istr >> player_roles[k];
            if (k!= (unsigned int) player_roles[k])
            {
                cout << "error: Wrong PType Number " << k << " " << player_roles[k] << endl;
            }
        }
        istr.clear();
        istr.str(getNextLine(is));
        for(unsigned int k=1; k<=num_player_types; k++) istr >> attraction_x[k];
        istr.clear();
        istr.str(getNextLine(is));
        for(unsigned int k=1; k<=num_player_types; k++) istr >> attraction_y[k];
        istr.clear();
        istr.str(getNextLine(is));
        for(unsigned int k=1; k<=num_player_types; k++) istr >> area_att_att[k];
        istr.clear();
        istr.str(getNextLine(is));
        for(unsigned int k=1; k<=num_player_types; k++) istr >> area_att_def[k];
        istr.clear();
        istr.str(getNextLine(is));
        for(unsigned int k=1; k<=num_player_types; k++) istr >> behind_ball[k];
        istr.clear();
        istr.str(getNextLine(is));
        for(unsigned int k=1; k<=num_player_types; k++) istr >> pos_min_x[k];
        istr.clear();
        istr.str(getNextLine(is));
        for(unsigned int k=1; k<=num_player_types; k++) istr >> pos_max_x[k];
        istr.clear();
        istr.str(getNextLine(is));
        for(unsigned int k=1; k<=num_player_types; k++) istr >> pos_min_y[k];
        istr.clear();
        istr.str(getNextLine(is));
        for(unsigned int k=1; k<=num_player_types; k++) istr >> pos_max_y[k];
        istr.clear();
        istr.str(getNextLine(is));
        for(unsigned int k=1; k<=num_player_types; k++) istr >> decision_algorithm[k];

        // Inneficient due to the way the data is stored
        for (unsigned int i = 1; i <= num_player_types; i++)
        {
            PlayerType* new_player_type = new PlayerType((PlayerType::PlayerRole) player_roles[i],
                                                         attraction_x[i], attraction_y[i],
                                                        area_att_att[i], area_att_def[i],
                                                        behind_ball[i],
                                                        pos_min_x[i], pos_max_x[i],
                                                        pos_min_y[i], pos_max_y[i],
                                                        (PlayerType::DecisionAlgorithm) decision_algorithm[i]);

            M_player_types.insert(PlayerTypes::value_type(player_roles[i], new_player_type));
        }

        return true;
    }


    bool Strategy::readFormations(istream& is, unsigned int num_formations)
    {
        cout << "Reading " << num_formations << " Formations" << endl;

        for (unsigned int i = 0; i < num_formations; i++)
        {
            istringstream istr(getNextLine(is));

            cout << istr.str() << endl;
            // Read formation id
            int formation_id = -1;
            istr >> formation_id;
            if ( formation_id == -1 ) {
                cerr << "ERROR reading formation id"
                          << formation_id
                          << istr.str()
                          << endl;
                return false;
            }

            // Read formation type
            int formation_type = -1;
            istr >> formation_type;
            if ( formation_type == -1 )
            {
                cerr << "Error reading formation type "
                          << formation_type
                          << istr.str()
                          << endl;
            }

            switch ( formation_type )
            {
                case formation::FT_SBSP:
                {
                    soccer::formation::Formation* new_formation_sbsp = new FormationSBSP(formation_id);
                    if ( !new_formation_sbsp->read(is) )
                    {
                        delete new_formation_sbsp;
                        return false;
                    }
                    M_formations.insert(Formations::value_type(new_formation_sbsp->id(),
                                                               new_formation_sbsp ));
                }
                break;
                case formation::FT_Delaunay:
                {
                    // Redifining non-virtual read method
                    istr.str(getNextLine(is));
                    string formation_file = "";
                    istr >> formation_file;
                    ifstream is_formation_dt(formation_file.c_str (), istream::in);
                    if (! is_formation_dt.is_open())
                    {
                        cerr << "FormationDT: Could not open formation file name "
                                 << formation_file
                                 << istr.str()
                                 << endl;
                        return false;
                    }

                    soccer::formation::FormationDT* new_formation_fcp_dt =
                                             new soccer::formation::FormationDT(formation_id);
                    if ( !new_formation_fcp_dt->read(is_formation_dt) )
                    {
                        delete new_formation_fcp_dt;
                        cerr << "FormationDT: Error reading delaunay formation"
                                  << formation_file
                                  << endl;
                        return false;
                    }
                    is_formation_dt.close();
                    new_formation_fcp_dt->readPlayerTypes(is);
                    M_formations.insert( Formations::value_type(new_formation_fcp_dt->id(), new_formation_fcp_dt ));
                }
                    break;
                default:
                {
                    cerr << "ERROR: Unknown formation type "
                              << endl
                              << istr.str()
                              << endl;
                    return false;
                }
            }
        }
        return true;
    }


    bool Strategy::readFluxMatrixes(istream& is, unsigned int num_flux_matrices)
    {
        cout << "Reading " << num_flux_matrices << " Flux Matrixes" << endl;

        for (unsigned int i=1; i <= num_flux_matrices; i++)
        {
            istringstream istr(getNextLine(is)); //Flux Matrixes Header
            int flux_id = -1;
            istr >> flux_id;
            if (i != (unsigned int) flux_id ) {
                cout << "error: Wrong Flux Number " << i << " " << flux_id << endl;
            }
            int type = -1;
            istr >> type;

            // Read flux file name
            istr.str(getNextLine(is));
            string flux_file;
            istr >> flux_file;

            if ( type == soccer::flux::FT_Matrix2D )
            {  //old 7x9 Flux
                FluxMatrix2D* flux_matrix_2d = new FluxMatrix2D();
                if ( !flux_matrix_2d->readFile(flux_file))
                {
                    delete flux_matrix_2d;
                    return false;
                }
                M_flux_matrices.insert(FluxMatrices::value_type(i, flux_matrix_2d ));
            }
            else if ( type == soccer::flux::FT_Delaunay )
            {
                FluxDT* flux_delaunay =
                        new FluxDT(rcsc::Rect2D(rcsc::Vector2D( -60.0, -45.0 ),
                                                Size2D( 120.0, 90.0 )));
                if ( !flux_delaunay->readFile(flux_file) )
                {
                    delete flux_delaunay;
                    return false;
                }
                M_flux_matrices.insert(FluxMatrices::value_type(i, flux_delaunay ));
            }
            else
            {
                cerr << "error: Unknown Flux Type" << type << endl;
                assert(false);
            }
        }
        return true;
    }


    bool Strategy::readSetplaySetsDefinitions(istream& is, unsigned int num_setplays_sets) {
        cout << "Reading " << num_setplays_sets << " Sets of Setplays" << endl;

        for (unsigned int i=1; i <= num_setplays_sets; i++)
        {
          istringstream istr(getNextLine(is));
          int plans_set_id = -1; //Set of Plans Number
          istr >> plans_set_id;
          if ( i != (unsigned int) plans_set_id) {
              cerr << "error: Wrong Set of Plans Number " << endl;
              return false;
          }

          int plans_set_type; //Type of Plans
          istr >> plans_set_type;

          istr.str(getNextLine(is)); // The next line holds the ids of the setplays that belong to the set

          SetplaySet* setplay_set = new SetplaySet();
          while ( !istr.eof() )
          {
              int setplay_id = -1;
              istr >> setplay_id;
              setplay_set->insert(setplay_id);
          }

          M_setplay_sets.insert(SetplaySets::value_type(i, setplay_set));
        }
        return true;
    }

    // Generic Strategy using Tactics, Formations, Player Types, FluxMatrixes and SetsofSetplays

    bool Strategy::read(const string& file_name)
    {
        ifstream ifs(file_name.c_str(), ifstream::in);

        if (!ifs.is_open())
        {
            cerr << "Strategy: Error opening stragey file " << file_name << endl;
            return false;
        }
        cout << "2D Strategy Configuration File OPEN!" << endl;

        // Strategy File Header and Strategy General Parameters
        istringstream istr(getNextLine(ifs));

        unsigned int num_tactics = 0;
        unsigned int num_formations = 0;
        unsigned int num_player_types = 0;
        unsigned int num_flux_matrices = 0;
        unsigned int num_setplay_sets = 0;

        istr >> num_tactics;
        istr >> num_formations;
        istr >> num_player_types;
        istr >> num_flux_matrices;
        istr >> num_setplay_sets;

        // Read time tactics
        istr.clear();
        istr.str(getNextLine(ifs));

        unsigned int num_time_tactics = 0;
        unsigned int num_opponent_tactics = 0;
        istr >> num_time_tactics;
        istr >> num_opponent_tactics;
        if (!readTimeTactics (ifs, num_time_tactics, num_opponent_tactics))
        {
            return false;
        }

        // Read other tactics
        if (!readTactics(ifs, num_tactics))
        {
            return false;
        }

        // Player Types Definition
        if (!readPlayerTypesDefinitions(ifs, num_player_types))
        {
            return false;
        }

        // Formation Definition
        if (!readFormations(ifs, num_formations))
        {
            return false;
        }

        // FLux Matrix Definitions
        if (!readFluxMatrixes(ifs, num_flux_matrices))
        {
            return false;
        }

        // Sets Of Setplays Definitions
        if (!readSetplaySetsDefinitions(ifs, num_setplay_sets))
        {
            return false;
        }

        // Set initial tactic
        // M_current_tactic = M_tactics.at(M_time_tactics.at(0)->tactic(ST_ME));
        // cout << "Setting initial tactic to #" << M_current_tactic->id() << endl;
        // ST_TimeTactic[1][0][3]); //Set tactic for game begin, generic opponent, draw

        print(std::cout);
        return true;
    }

    /****************************************************************************************/
    /********************************* Formation Analysis ***********************************/
    /****************************************************************************************/

    // Why not use situations? (see [Reis and Lau 2001]) What should be the tactic and formation
    // for each situation? How to define different player types?

    bool Strategy::updateTactic( const IWorldModel* world_model )
    {
        Tactic* next_tactic = NULL;
        unsigned int next_tactic_id = M_tactic_selector->getTactic( world_model );
        if ( M_tactics.find(next_tactic_id) != M_tactics.end() )
        {
            next_tactic = M_tactics.at( next_tactic_id );
        }

        // Change tactic only if it is different than the current one
        if ( M_current_tactic == next_tactic )
        {
            return false;
        }

        M_current_tactic = next_tactic;
        tacticChanged( );

        // Choose the flux used in this tactic
        // Futur flux selection be based on opponent modeling and or simply the score
        // M_current_flux = M_flux_matrices.at(M_current_tactic->fluxMatrix());

        return true;


        cerr << __FILE__ << ":" << __FUNCTION__ << ":" << __LINE__
             << " No time tactic available for selection" << endl;
        return false;
    }

    bool Strategy::updateFormation( const IWorldModel* world_model )
    {
        assert ( M_current_tactic != NULL );
            // " No tactic from which to choose the new formation"

        const TacticWithSituations& current_tactic = static_cast<const TacticWithSituations&>(*M_current_tactic);

        GameSituation game_situation = world_model->currentGameMode().gameSituation();
        unsigned int next_formation_id = current_tactic.formation( game_situation );

        soccer::formation::Formation* next_formation = M_formations.at( next_formation_id );

        if ( M_current_formation ==  next_formation )
        {
            return false;
        }

        M_current_formation = next_formation;
        formationChanged( );

        return true;
    }

    /****************************************************************************************/
    /********************** Strategy - Tactical Decision and Change  ************************/
    /****************************************************************************************/


    bool Strategy::updatePositions(  const IWorldModel* world_model )
    {
        assert( M_current_formation != NULL );

        for (UnumPositions::const_iterator itp = M_positions.begin();
             itp  != M_positions.end();
             itp ++)
        {
            // If a plan is in execution should decide other positioning for pla
            if ( decidePlayerPositionAccordingToPlan( itp ->first ))
            { // If true, it means that the player should have already got a position for him
                continue;
            }

            rcsc::Vector2D player_position = M_current_formation->position(itp ->first, *world_model);

            stringstream debug;
            debug << " # " << itp ->first;
            debug << " PrevSPos" << M_positions[itp ->first];
            debug << " CurPos" << world_model->playerPosition( TA_OUR, itp->first ) ;
            debug << " FormPos" << player_position;

            // Check if there was a change between formations and the transition period is "active"
            /*if ( M_formations_config.smoothTransitions() && M_last_formation != NULL )
            {
                // +1 => Apply transition immediately at the start of transition (and not at the last
                unsigned int transition_time_elapsed = world_model->gameTime() - M_last_formation_change + 1;
                if (transition_time_elapsed < M_formations_config.smoothTransitionsDelay())
                {
                    // Point past_position = M_last_formation->position( it_player->first, *world_model);
                    // DO NOT USE FORMATION POSITION AS ERROR MARGIN CAN BE VERY BIG
                    // PLAYER SHOULD ADJUST VELOCITY ACCORDING TO TRUNCATED POSITION
                    Point past_position = world_model->ourPlayerPosition(itp ->first);
                    Point adjust_position = past_position;
                    Point to_new_position(player_position.x - past_position.x,
                                             player_position.y - past_position.y);
                    to_new_position.scale( (float) transition_time_elapsed /
                                            M_formations_config.smoothTransitionsDelay());
                    // Set new player position based on the 2 formations positions and the transition elapsed time
                    adjust_position += to_new_position;

                    debug << " FormTrans " << M_last_formation->id()
                       << "->" << M_current_formation->id()
                       << " at " << 100 * transition_time_elapsed /
                                    (float) M_formations_config.smoothTransitionsDelay() << " \%"
                       << " Adjust" << adjust_position
                       << " Final" << player_position
                       << std::endl;
                    player_position = adjust_position;
                }
            }*/

            PlayerType& player_type = playerType(itp ->first);

            player_position = positionLeagueSpecific(player_position, world_model);
            debug << " PLeagueSpecific" << player_position;

            // Modify player position after formation
            /*if ( player_type )
            {*/
                debug << " PlayerType "  << player_type.role();
    ;                    // Player Attraction => HACK Ignoring attraction
                /*player_position.x +=  world_model->ballPosition().x * player_type->attractionX();
                player_position.y +=  world_model->ballPosition().y * player_type->attractionY();
                dlog.addText( Logger::POSITIONING,
                             __FILE__ " SPos after PTypeAttraction(%.2f,%.2f)",
                             player_position.x, player_position.y);
                player_position = positionAreaAttraction(player_position, world_model->ballPosition());
                dlog.addText( Logger::POSITIONING,
                             __FILE__ " SPos after PTypeAttractionArea(%.2f,%.2f)",
                             player_position.x, player_position.y);*/
                player_position = positionMaxMin(player_type, player_position, world_model);
                debug << " PMaxMin" << player_position;
            //}
            dlog.addText( Logger::POSITIONING, __FILE__ " %s", debug.str().c_str());
            dlog.addCircle( Logger::POSITIONING, rcsc::Circle2D(player_position, 1), "blue", false);
            M_positions[itp->first] = player_position;
        }
        return true;
    }

    // Improve this using situations, attractions by points or lines in some situations, etc.

    rcsc::Vector2D Strategy::positionAreaAttraction(const rcsc::Vector2D& player_position,
                                                    const rcsc::Vector2D& ball_position,
                                                    const IWorldModel *world_model)
    {
        rcsc::Vector2D new_player_position(player_position);
        double field_half_width = world_model->field().size().width() / 2;
        double field_quarter_width = field_half_width / 2;
        double field_quarter_length = world_model->field().size().length() / 4;

        // attraction is less near the sides of the field
        if (new_player_position.y > field_quarter_width)
        {
            new_player_position.y = new_player_position.y - (new_player_position.y-field_quarter_width) * 0.25;
        }
        if (new_player_position.y < -field_quarter_width)
        {
            new_player_position.y = new_player_position.y - (new_player_position.y+field_quarter_width) * 0.25;
        }

        // players converge to the center (area) near the own/opponent goals
        if (new_player_position.x > field_quarter_length)
        {
            double factor1 = (new_player_position.x - field_quarter_length)/(field_quarter_length);
            double factor2 = 1.0 - fabs(ball_position.y/field_half_width);
            new_player_position.y = new_player_position.y * (1.0-factor1*0.5 - factor2*0.3);
        }

        if (new_player_position.x < - field_quarter_length)
        {
            double factor1 = (new_player_position.x + field_quarter_length)/(-field_quarter_length);
            double factor2 = 1.0 - fabs(ball_position.y/field_half_width);
            new_player_position.y = new_player_position.y * (1.0-factor1*0.4-factor2*0.3);
        }
        return new_player_position;
    }


    rcsc::Vector2D Strategy::positionMaxMin(const PlayerType& player_type,
                                            const Vector2D &player_position,
                                            const IWorldModel* world_model)
    {
        rcsc::Vector2D new_player_position(player_position);
        double field_half_length = world_model->field().size().length() / 2;
        double field_half_width = world_model->field().size().width() / 2;

        // SBSP inside the field and respecting Max and Min x and y conditions
        if (new_player_position.x < -0.99 * field_half_length)
            new_player_position.x = -0.99 * field_half_length;
        if (new_player_position.y < -0.99 * field_half_width)
            new_player_position.y = -0.99 * field_half_width;
        if (new_player_position.y > 0.99 * field_half_width)
            new_player_position.y = 0.99 * field_half_width;
        if (new_player_position.x > player_type.maxPosX())
            new_player_position.x = player_type.maxPosX();
        if (new_player_position.x < player_type.minPosX())
            new_player_position.x = player_type.minPosX();
        if (new_player_position.y > player_type.maxPosY())
            new_player_position.y = player_type.maxPosY();
        if (new_player_position.y < player_type.minPosY())
            new_player_position.y = player_type.minPosY();

        return new_player_position;
    }

    /*** ACCESSORS ***/


    PlayerType &Strategy::playerType(const soccer::Unum &number) const
    {
        assert ( M_current_formation != NULL );

        return *M_player_types.at(M_current_formation->playerType(number));
    }


    const Strategy::SetplaySet&
    Strategy::availableSetplays() const
    {
        assert ( M_current_tactic != NULL && !M_setplay_sets.empty() );
        /*if (SetplayConfig::instance().useSetplays() )
        {
            cerr << __FILE__ << __FUNCTION__ << ":" << __LINE__
                 << " Use of setplays is disabled" << endl;
            return NULL;
        }

        if ( ! TacticConfig::instance().useTactics())
        {
            cerr << __FILE__ << __FUNCTION__ << ":" << __LINE__
                 << " Use of tactics is disabled, returning no setplays set" << endl;
            return NULL; // TODO Should return all setplay sets
        }

        if ( M_setplay_sets.empty() )
        {
            cerr << __FILE__ << __FUNCTION__ << ":" << __LINE__
                 << " There are no setplays sets" << endl;
            return NULL;
        }*/

        assert ( M_setplay_sets.count( M_current_tactic->setplaysSet()) != 0 );
        /*{
            cerr << __FILE__ << __FUNCTION__ << ":" << __LINE__
                 << " There is no setplay set nr. " << M_current_tactic->setplaysSet()
                 << " for tactic " << M_current_tactic->id() << endl;
            return NULL;
        }*/

        return *M_setplay_sets.at( M_current_tactic->setplaysSet() );
    }

    /****************************************************************************************/
    /******************** DPRE - Dynamic Positioning and Role Exchange **********************/
    /****************************************************************************************/

    /*
    void Strategy::reset()
    {
        // What to do with player positions? => Stay in formation?
        if ( M_dpre->active() && StrategyConfig::instance().useDPRE() )
        {
            M_dpre->reset();
        }
        // Change tactic to NULL or default
    }*/


    const Flux*
    Strategy::currentFlux() const
    {
        assert ( M_current_flux != NULL );

        return M_current_flux;
    }


    const Tactic* Strategy::currentTactic() const
    {
        assert ( M_current_tactic != NULL );

        return M_current_tactic;
    }


    const Strategy::SetplaySet*
    Strategy::currentSetplaySet() const
    {
        return M_setplay_sets.at( currentTactic()->setplaysSet() );
    }


    const soccer::formation::Formation*
    Strategy::currentFormation( ) const
    {
        assert ( M_current_formation != NULL );

        return M_current_formation;
    }

    const soccer::formation::Formation*
    Strategy::lastFormation( ) const
    {
        assert ( M_last_formation != NULL );

        return M_last_formation;
    }


    bool Strategy::isSetplayUsed(unsigned int id) const
    {
        if ( !SetplayConfig::instance().useSetplays() )
        {
            cerr << __FILE__ << ":" << __LINE__ << __FUNCTION__
                 << " Setplays are configured not to be used " << endl;
            return false;
        }

        assert ( M_current_tactic != NULL );

        const SetplaySet& tactic_setplays =  *M_setplay_sets.at( M_current_tactic->setplaysSet() );

        return find( tactic_setplays.begin(), tactic_setplays.end(), abs(id) ) != tactic_setplays.end();
    }

    /*****************/
    /* SANITY CHECKS */
    /*****************/


    bool Strategy::sanityCheck() const
    {
        bool result = true;
        // Sanity check for all tactics
        for (Tactics::const_iterator it_tactic = M_tactics.begin();
             it_tactic != M_tactics.end();
             it_tactic++)
        {
            result &= sanityCheck( *(*it_tactic).second );
        }

        // Sanity check for all formations
        for (typename Formations::const_iterator it_formation = M_formations.begin();
             it_formation != M_formations.end();
             it_formation++)
        {
            // TODO Check ids!
            result &= sanityCheck( *(*it_formation).second );
        }

        return result;
    }


    bool Strategy::sanityCheck(const Tactic& tactic) const
    {
        bool result = true;

        // Check if default formation specified by tactic exists
        if ( M_formations.find(tactic.defaultFormation()) == M_formations.end() )
        {
            cerr << __FILE__ << ":" << __LINE__ << __FUNCTION__
                 << " Formation #" << tactic.defaultFormation()
                 << " specified in tactic " << tactic.id()
                 << " does not exist" << endl;
            result = false;
        }

        // Check if flux matrix formation specified by tactic exists
        if ( M_flux_matrices.find(tactic.flux()) == M_flux_matrices.end() )
        {
            cerr << __FILE__ << ":" << __LINE__ << __FUNCTION__
                 << " Flux #" << tactic.flux() << " specified in tactic " << tactic.id()
                 << " does not exist" << endl;
            result = false;
        }

        // Check if set of setplays specified by tactic exists
        if ( M_setplay_sets.find(tactic.setplaysSet()) == M_setplay_sets.end() )
        {
            cerr << __FILE__ << ":" << __LINE__ << __FUNCTION__
                 << " Setplay set #" << tactic.setplaysSet()
                 << " specified in tactic " << tactic.id()
                 << " does not exist" << endl;
            result = false;
        }

        switch (tactic.type())
        {
            case Tactic::TacticWithSituations:
            {
            // Check if all formations for game situations exist
                const TacticWithSituations& tactic_with_situations = static_cast<const TacticWithSituations&>(tactic);
                for (TacticWithSituations::GameSituationFormations::const_iterator it_formation = tactic_with_situations.gameSituationFormations().begin();
                     it_formation != tactic_with_situations.gameSituationFormations().end();
                     it_formation++)
                {
                    unsigned int formation_id = it_formation->second;
                    GameSituation game_situation = it_formation->first;
                    if ( M_formations.find(formation_id) == M_formations.end() )
                    {
                        cerr << __FILE__ << ":" << __LINE__ << __FUNCTION__
                             << " Formation #" << formation_id
                             << " for situation " << game_situation
                             << " specified in tactic " << tactic.id()
                             << " does not exist" << endl;
                        result = false;
                    }
                }
                break;
            }
        }

        return result;
    }


    bool Strategy::sanityCheck(const soccer::formation::Formation& formation) const
    {
        bool result = true;

        // Check if the player types specified for the formation all exist
        for ( soccer::formation::Formation::PlayerTypes::const_iterator it_pltype
             = formation.playerTypes().begin();
              it_pltype != formation.playerTypes().end();
              it_pltype++)
        {
            if ( M_player_types.find( it_pltype->second ) == M_player_types.end() )
            {
                cerr << __FILE__ << ":" << __LINE__ << __FUNCTION__
                     << " Player type #" << it_pltype->second
                     << " in formation " << formation.id()
                     << " does not exist" << endl;
                result = false;
            }
        }

        /*switch ( formation.type() )
        {
            case Formation::Delaunay:
            {
                const FormationDT& formation_dt = static_cast<const FormationDT&>(formation);
                formation_dt.print(cout);
                cout << endl;
                break;
            }
            case Formation::SBSP:
            {
                const FormationSBSP& formation_sbsp = static_cast<const FormationSBSP&>(formation);
                break;
            }
        }*/

        return result;
    }


    std::ostream& Strategy::print(std::ostream& os) const
    {
        // Print player types
        os << "-- Player Types --" << std::endl;
        for ( PlayerTypes::const_iterator itpt = M_player_types.begin(); itpt != M_player_types.end(); itpt++ )
        {
            os << "Player Type " << itpt->first << std::endl;
            itpt->second->print(os);
        }
        return os;
    }


}

}
