#ifndef FCPORTUGAL_STRATEGY_H
#define FCPORTUGAL_STRATEGY_H

#include <fstream>
#include <vector>
#include <set>

#include <soccer/iworld_model.h>

#include <rcsc/common/logger.h>
#include <rcsc/geom/vector_2d.h>
#include <rcsc/geom/ray_2d.h>

#include <strategy/flux/flux.h>
#include <strategy/formation/formation.h>
#include <strategy/tactic/tactic.h>
#include <strategy/tactic/time_tactic.h>
#include <strategy/player_type.h>

#include <strategy/dpre.h>

#include <strategy/util.h>

using namespace std;
using namespace rcsc;

namespace soccer
{
    namespace strategy
    {

        class Strategy
        {
        public:
            typedef std::set<unsigned int> SetplaySet;

            Strategy();
            virtual ~Strategy();

            /*!
              \brief Get the strategic position for a given player number
              \param player_unum The number of the player whose strategic position to get
              \return The strategic position of the player number
              */
            const rcsc::Vector2D& position(const soccer::Unum& number) const;

            bool init( const std::string& strategy_config_file );

            void reset(void);

            const SetplaySet &availableSetplays() const;

            // Accessors
            const tactic::Tactic* currentTactic( ) const;
            const SetplaySet* currentSetplaySet() const;
            const formation::Formation* currentFormation() const;
            const formation::Formation* lastFormation() const;
            const flux::Flux* currentFlux() const;
            bool isSetplayUsed(unsigned int id) const;

            // Sanity checksfor strategy data read
            bool sanityCheck() const;
            bool sanityCheck(const tactic::Tactic& tactic) const;
            bool sanityCheck(const formation::Formation& formation) const;

            /*!
              \brief Print out the current configurations of the strategy instance
              */
            std::ostream& print(std::ostream& os) const;
        protected:

            virtual bool decidePlayerPositionAccordingToPlan(unsigned int player_number) = 0;

            // Should try to assess tactical behavior somehow
            // Register times at which tactics are changed
            tactic::Tactic* M_current_tactic;
            formation::Formation* M_current_formation;
            formation::Formation* M_last_formation;
            flux::Flux* M_current_flux;

            // Track changes in formation to handle smooth transitions between position
            // unsigned int M_last_formation_change; // DANGER This does not consider stop time

            // Assuming all identifiers in strategy are sequential and are in the range 1..N for all concepts
            typedef std::map<unsigned int, SetplaySet* > SetplaySets;

            // Player strategic positions at a given time
            formation::Formations M_formations;
            PlayerTypes M_player_types;
            flux::FluxMatrices M_flux_matrices;
            tactic::Tactics M_tactics;
            tactic::ITacticSelector* M_tactic_selector;

            // Setplays are managed by a different instance (SetplayManager in Planner)
            SetplaySets M_setplay_sets;

            // Dynamic Positioning and Role Exchange - Players may exchange their positionings and player
            // types in the current formation if the utility of that exchange is positive for the team

            // Strategy update
            void update(const IWorldModel* world_model);
            //Decides the tactic based on time, result an opponent
            bool updateTactic( const IWorldModel* world_model  );
            //Decides the formation based on the tactic and current game situation
            bool updateFormation(  const IWorldModel* world_model );

            virtual bool updatePositions( const IWorldModel* world_model );

            virtual void tacticChanged( ) = 0;
            virtual void formationChanged( ) = 0;

            rcsc::Vector2D positionAreaAttraction(const rcsc::Vector2D& player_position,
                                                  const rcsc::Vector2D& ball_position,
                                                  const IWorldModel *world_model);
            rcsc::Vector2D positionMaxMin(const PlayerType& player_type,
                                          const rcsc::Vector2D& player_position,
                                          const IWorldModel *world_model);

            virtual rcsc::Vector2D positionLeagueSpecific(rcsc::Vector2D player_pos,
                                                          const IWorldModel* world_model) = 0;

            // Returns Player Types Characteristics: Attraction, BehindBall, Decision Algorithm
            // MinX, MaxX, MinY or MaxY of player t
            PlayerType& playerType(const soccer::Unum& number) const;

            // Helper functions to read strategy configuration file
            bool read(const std::string& file_name);
            bool readTactics(std::istream& is, unsigned int num_tactics);
            bool readTimeTactics(std::istream& is,
                                 unsigned int num_time_tactics,
                                 unsigned int num_opponent_tactics);
            bool readPlayerTypesDefinitions(std::istream& is, unsigned int num_player_types);
            bool readFormations(std::istream& is, unsigned int num_formations);
            bool readFluxMatrixes(std::istream& is, unsigned int num_flux_matrices);
            bool readSetplaySetsDefinitions(std::istream& is, unsigned int num_setplay_sets);

        protected:
            UnumPositions M_positions;

        };

    }
}
#endif // FCPORTUGAL_STRATEGY_H
